unit Fields;

interface

uses
  Globals, Rasterfield;

type TField = class
  private
    Data                : array of TNode;  // Gitternetzarray
    SizeX, SizeY, SizeQ : integer;         // Feldgren

    function GetFieldPtr(X, Y: integer): PNode;             // read FieldPtr
    function GetField(X, Y: integer)   : TNode;             // read Field
    procedure SetField(X, Y: integer; const Value: TNode);  // write Field
    function GetDataSize : integer;  // read DataSize
    function GetDataPtr : pointer;  // read DataPtr
  public
    constructor Create(Raster: TRaster); // Konstruktor
    destructor Destroy; override;        // Destruktor

    // EIGENSCHAFTEN
    property Field[X, Y: integer]   : TNode read GetField
                                            write SetField; default;
      // das eigentliche Spielfeld
    property FieldPtr[X, Y: integer]: PNode read GetFieldPtr;
      // Zeiger auf die Spielfeldknoten (fr schnellen wahlfreien Zugriff)
    property Width : integer read SizeX;  // Spielfeldbreite
    property Height: integer read SizeY;  // Spielfeldhhe
    property DataSize: integer read GetDataSize;  // Gre der Daten
    property DataPtr: pointer read GetDataPtr;  // Zeiger auf die Daten

    // METHODEN
    procedure DeterminePossibleOwners; overload;
      // ganzes Feld prfen
    procedure DeterminePossibleOwners(LastMove: TMove); overload;
      // Teilfeld prfen
    function MoveCheck(Move:TMove; Flags:TMoveCheckFlags): boolean;
      // Zugtest / Zug durchfhren -- HardTest gibt an,
      // ob "Strichverlngerungen" mit geprft werden sollen
    function MayBeOccupiedBy(Player, X, Y: integer): boolean;
      // prft, ob ein Feld von einem bestimmten Spieler besetzt
      // werden knnte
  end;

implementation

uses
  Tables;

constructor TField.Create(Raster: TRaster);
var X, Y           : integer;
    A1, A2, A3, A4 : boolean;
begin
  SizeX := Raster.Width - 1;   // Rasterbreite-1
  SizeY := Raster.Height - 1;  // Rasterhhe-1
  SizeQ := SizeX * SizeY;
  SetLength(Data,SizeQ);

  for Y:=0 to (SizeY - 1) do
    for X:=0 to (SizeX - 1) do
      with Data[Y*SizeX + X] do begin
        PossibleDirs := 0;
        A1 := Raster[X, Y];      // Rasterfeld links oben
        A2 := Raster[X+1, Y];    // Rasterfeld rechts oben
        A3 := Raster[X+1, Y+1];  // Rasterfeld rechts unten
        A4 := Raster[X, Y+1];    // Rasterfeld links unten
        if ((A1 OR A4) AND (A2 OR A3)) then  // rechts und links -> -
          PossibleDirs := PossibleDirs OR DIR_H;
        if ((A1 OR A2) AND (A3 OR A4)) then  // oben und unten -> |
          PossibleDirs := PossibleDirs OR DIR_V;
        if (A2 AND A4) then                  // re. ob. und li. unt. -> /
          PossibleDirs := PossibleDirs OR DIR_D;
        if (A1 AND A3) then                  // li. ob. und re. unt -> \
          PossibleDirs := PossibleDirs OR DIR_A;
        if (A1 OR A2 OR A3 OR A4) then begin
          Owner := OWNER_EMPTY;  // noch kein Besitzer
          Dirs  := 0;            // zu Beginn keine Linien gesetzt
          PossibleOwners := OWNER_UNDECIDED;
        end else begin
          Owner := OWNER_OOF;    // Knoten auerhalb des Spielfeldes
          Dirs  := 0;
          PossibleOwners := OWNER_OOF;
        end;
      end;
end;

destructor TField.Destroy;
begin
  Finalize(Data);
  inherited Destroy;
end;

function TField.GetField(X, Y: integer): TNode;
begin
  if (X >= 0) AND (Y >= 0) AND (X < SizeX) AND (Y < SizeY) then
    Result := Data[Y * SizeX + X]
  else with Result do begin
    Owner        := OWNER_OOF;
    PossibleDirs := 0;
    Dirs         := -1;
  end;
end;

procedure TField.SetField(X, Y: integer; const Value: TNode);
begin
  if (X >= 0) AND (Y >= 0) AND (X < SizeX) AND (Y < SizeY) then
    Data[Y * SizeX + X] := Value;
end;

function TField.GetFieldPtr(X, Y: integer): PNode;
begin
  if (X >= 0) AND (Y >= 0) AND (X < SizeX) AND (Y < SizeY)
    then Result := @Data[Y * SizeX + X]
    else Result := nil;
end;

function TField.GetDataSize: integer;
begin
  Result := SizeQ * sizeof(TNode);
end;

function TField.GetDataPtr: pointer;
begin
  Result := @Data[0];
end;

procedure TField.DeterminePossibleOwners;
var X, Y: integer;
    M0  : boolean;
    PO  : integer;
begin
  for Y := 0 to (SizeY - 1) do
    for X := 0 to (SizeX - 1) do begin
      M0 := MayBeOccupiedBy(0, X, Y);
      if MayBeOccupiedBy(1, X, Y) then begin
        if M0 then PO := OWNER_UNDECIDED
        else PO := 1;
      end else begin
        if M0 then PO := 0
        else PO := OWNER_DEAD;
      end;
      FieldPtr[X, Y]^.PossibleOwners := PO;
    end;
end;

procedure TField.DeterminePossibleOwners(LastMove: TMove);
var X, Y: integer;
    M0  : boolean;
    PO  : integer;
begin
  for Y := (LastMove.Y - 3) to (LastMove.Y + 3) do
    if (Y >= 0) and (Y <= SizeY - 1) then
      for X := (LastMove.X - 3) to (LastMove.X + 3) do
        if (X >= 0) and (X <= SizeX - 1) then begin
          M0 := MayBeOccupiedBy(0, X, Y);
          if MayBeOccupiedBy(1, X, Y) then begin
            if M0 then begin
              PO := OWNER_UNDECIDED;
            end else
              PO := 1;
          end else begin
            if M0 then
              PO := 0
            else
              PO := OWNER_DEAD;
          end;
          FieldPtr[X, Y]^.PossibleOwners := PO;
        end;
end;

function TField.MoveCheck(Move: TMove; Flags: TMoveCheckFlags): boolean;
var ND0      : PNode;  // zu belegendes Feld
    ND1, ND2 : PNode;  // umliegende Felder
  function RejectAdjacentLines(X, Y, DirMask: integer): boolean;
  begin
    with Field[X,Y] do
      Result := (Owner <> OWNER_EMPTY) AND (Owner <> OWNER_OOF) AND
                (Owner <> Move.Player) AND ((Dirs AND DirMask) <> 0);
  end;
begin
  Result := false;

  // ungltige Koordinaten ablehnen (klar)
  if (Move.X < 0) OR (Move.Y < 0) OR (Move.X >= SizeX) OR (Move.Y >= SizeY)
    then exit;

  // Zielfeld laden
  ND0 := FieldPtr[Move.X,Move.Y];

  // Ablehnen, wenn Zug in diese Richtung nicht mglich ist
  if (ND0^.PossibleDirs AND Move.Dir) = 0 then exit;

  // umliegende Felder ND1 und ND2 laden
  case Move.Dir of
    DIR_H : begin
              ND1 := FieldPtr[Move.X-1,Move.Y];
              ND2 := FieldPtr[Move.X+1,Move.Y];
            end;
    DIR_V : begin
              ND1 := FieldPtr[Move.X,Move.Y-1];
              ND2 := FieldPtr[Move.X,Move.Y+1];
            end;
    DIR_D : begin
              ND1 := FieldPtr[Move.X-1,Move.Y+1];
              ND2 := FieldPtr[Move.X+1,Move.Y-1];
            end;
    DIR_A : begin
              ND1 := FieldPtr[Move.X-1,Move.Y-1];
              ND2 := FieldPtr[Move.X+1,Move.Y+1];
            end;
  else
    exit;  // ungltige Richtung
  end;

  // Ablehnen, wenn umliegende Felder einem Gegner gehren
  if (ND0 = nil) OR
    ((ND0^.Owner <> OWNER_EMPTY) AND (ND0^.Owner <> Move.Player)) then exit;
  if (ND1 = nil) OR
    ((ND1^.Owner <> OWNER_EMPTY) AND (ND1^.Owner <> Move.Player)) then exit;
  if (ND2 = nil) OR
    ((ND2^.Owner <> OWNER_EMPTY) AND (ND2^.Owner <> Move.Player)) then exit;

  // Wenn Hardtest, dann Ablehnen, wenn Linie ergnzt werden soll
  // (d.h. umliegende Felder eine Linie in die selbe Richtung wie die zu
  // zeichnende enthalten)
  if ((ND0^.Dirs AND Move.Dir) <> 0) then exit;
  if not(mcSoftTest IN Flags) then
    if ((ND1^.Dirs AND Move.Dir) <> 0) OR
       ((ND2^.Dirs AND Move.Dir) <> 0) then exit;

  // brige Flle abprfen
  case Move.Dir of
    DIR_D : if RejectAdjacentLines(Move.X-1 ,Move.Y  , DIR_A) OR
               RejectAdjacentLines(Move.X   ,Move.Y-1, DIR_A) OR
               RejectAdjacentLines(Move.X+1 ,Move.Y  , DIR_A) OR
               RejectAdjacentLines(Move.X   ,Move.Y+1, DIR_A)
            then exit;
    DIR_A : if RejectAdjacentLines(Move.X-1 ,Move.Y  , DIR_D) OR
               RejectAdjacentLines(Move.X   ,Move.Y-1, DIR_D) OR
               RejectAdjacentLines(Move.X+1 ,Move.Y  , DIR_D) OR
               RejectAdjacentLines(Move.X   ,Move.Y+1, DIR_D)
            then exit;
  end;

  Result := true;

  if not(mcCommit IN Flags) then exit;
  with ND0^ do begin
    Dirs           := Dirs OR Move.Dir;
    PossibleDirs   := PossibleDirs - Move.Dir;
    Owner          := Move.Player;
    PossibleOwners := Move.Player;
  end;
  with ND1^ do begin Owner := Move.Player; PossibleOwners := Move.Player; end;
  with ND2^ do begin Owner := Move.Player; PossibleOwners := Move.Player; end;
end;

function TField.MayBeOccupiedBy(Player, X, Y : integer) : boolean;
var Move: TMove;
    i: integer;
begin
  Result := false;  // eine pessimistische Haltung ist mitunter ntzlich :)
  if FieldPtr[X, Y]^.Owner = Player then Result := true;
    //was tun, wenn schon besetzt?
  if FieldPtr[X, Y]^.Owner <> OWNER_EMPTY then exit;
  // Feld ist leer -> prfen, ob es einen gltigen Zug gibt, mit dem der Spieler
  // dieses Feld besetzen kann
  Move.X := X;
  Move.Y := Y;
  Move.Player := Player;
  for i := 1 to LINES_SURROUNDING do
    if MoveCheck(AddDelta(Move, DeltaTable[1].Items[i]), []) then begin
      Result:=true;
      exit;
    end;
end;

end.
