unit Player;

interface

uses
  Fields, Rasterfield, Globals;

type TPlayer = class
               private
                 ThisPlayer      : integer;
                   // Identifikator
                 FRemainingField : TField;
                   // Referenzfeld fr die Aktualisierung von RemainingMoves;
                   // enthlt alle Linien, die NICHT mehr gezogen werden knnen
                 FRemainingMoves : integer;
                   // Anzahl der aktuell noch mglichen Zge
                 FIsHuman        : boolean;
                   // KI- oder Human-Player
                 FAILevel        : integer;
                   // KI-Level
                 FName           : string;
               public
                 constructor Create(Id: integer; Raster: TRaster;
                                    _Name: string; Human: boolean;
                                    _AILevel:integer);
                   // Konstruktor
                 destructor  Destroy; override;
                   // Destruktor

                 // EIGENSCHAFTEN
                 property RemainingField : TField  read FRemainingField; {DEBUG}
                 property RemainingMoves : integer
                   read FRemainingMoves write FRemainingMoves; {DEBUG}
                 property IsHuman        : boolean read FIsHuman;
                 property Name           : string  read FName;
                 property AILevel        : integer read FAILevel write FAILevel;

                 // METHODEN
                 function DetermineBlockedMoves(Move: TMove;
                                                Flags: TDBMFlags): integer;
                   // RemainingField mit Move neu berechnen, falls dbmCommit
                   // und gibt Anzahl der durch Move fr diesen Spieler
                   // blockierten Zge zurck
                 function DetermineSafedNodes(Move: TMove): integer;
                   // Anzahl der durch Move neu hinzugekommenen fr diesen
                   // Spieler gesicherten Gitternetzknoten
                 function DetermineMove: TMove;
                   // Zug je nach KI-Level bestimmen
                 function DetermineRandomMove: TMove;
                   // zuflligen Zug bestimmen
                 function DetermineRandomBestOffensiveMove: TMove;
                   // Zug mit maximaler Bewertung bestimmen
                 function DetermineRandomBestDefensiveMove: TMove;
                   // Zug mit den meisten gesicherten Gitterpunkten bestimmen
                 function DetermineRandomBestAverageMove: TMove;
                   // Kombination der beiden vorhergehenden Strategien
                 function DetermineBestMiniMaxMove: TMove;
                   // Zug mit MinMax-Algorithmus bestimmen
                 function DetermineBestAlphaBetaMove: TMove;
                   // Zug mit MinMax-Alg. und Alpha-Beta-Heuristik bestimmen
                 function CanMove: boolean;
                   // prft, ob der Spieler noch einen Zug ausfhren kann
                 procedure DetermineBestRatedMoves(var Count: integer;
                                                   var RatedMoves: TRatedMoves);
               end;

implementation

uses
  Game, Tables, Progress, SysUtils, Math,
  {Debug} Main;

constructor TPlayer.Create(Id: integer; Raster: TRaster;
                           _Name: string; Human: boolean; _AILevel:integer);
var X,Y : integer;
begin
  inherited Create;
  ThisPlayer := Id;
  FIsHuman   := Human;
  FName      := _Name;
  FAILevel   := _AILevel;
  FRemainingField := TField.Create(Raster);
  FRemainingMoves := 0;
  for Y := 0 to (FRemainingField.Height - 1) do
    for X := 0 to (FRemainingField.Width - 1) do
      with RemainingField[X,Y] do begin
        if (PossibleDirs and DIR_H) <> 0 then inc(FRemainingMoves);
        if (PossibleDirs and DIR_V) <> 0 then inc(FRemainingMoves);
        if (PossibleDirs and DIR_D) <> 0 then inc(FRemainingMoves);
        if (PossibleDirs and DIR_A) <> 0 then inc(FRemainingMoves);
      end;
end;

destructor TPlayer.Destroy;
begin
  RemainingField.Free;
  inherited Destroy;
end;

function TPlayer.DetermineBlockedMoves(Move: TMove; Flags: TDBMFlags): integer;
var TmpMove  : TMove;
    mcFlags  : TMoveCheckFlags;
    i        : integer;

  procedure CheckThisMove(ThisMove: TMove);
  begin
    if RemainingField.MoveCheck(ThisMove,mcFlags) then begin
      if dbmCommit in Flags then dec(FRemainingMoves);
      inc(Result);
    end;
  end;

begin
  Result  := 0;
  TmpMove := Move;
  if dbmCommit in Flags then
    mcFlags := [mcSoftTest, mcCommit]
  else
    mcFlags := [mcSoftTest];

  // Spieler hat den Strich selbst gezogen --> diesen und die zwei mglichen
  // Verlngerungen eintragen
  if Move.Player = ThisPlayer then begin
    for i := 1 to LINES_EXTENDING do
      CheckThisMove(AddDelta(Move,DeltaTable[Move.Dir].Items[i]));
    exit;
  end;

  // Spieler hat den Strich nicht selbst gezogen --> alle mglichen eigenen
  // dadurch verhinderten Zge eintragen
  Move.Player := ThisPlayer;
  for i := 1 to DeltaTable[Move.Dir].Count do begin
    TmpMove  := AddDelta(Move, DeltaTable[Move.Dir].Items[i]);
    if DoMovesCross(TmpMove, Move) then CheckThisMove(TmpMove);
  end;
end;

function TPlayer.DetermineSafedNodes(Move: TMove) : integer;
var TestX, TestY, i  : integer;
    TmpMove          : TMove;
    ByP1, ByP2       : boolean;
    ND  : array [1..LINES_EXTENDING] of TNode;
    PND : array [1..LINES_EXTENDING] of PNode;
begin
  Result := 0;

  with TheGame.Field do begin
    for i := 1 to 3 do begin
      TmpMove := AddDelta(Move, DeltaTable[Move.Dir].Items[i]);
      PND[i]  := FieldPtr[TmpMove.X,TmpMove.Y];
      ND[i]   := PND[i]^;
    end;
    MoveCheck(Move,[mcCommit]);

    for TestX := Move.X - 3 to Move.X + 3 do
      if (TestX >= 0) and (TestX <= Width - 1) then
        for TestY := Move.Y - 3 to Move.Y + 3 do
          if (TestY >= 0) and (TestY <= Height - 1) then begin
            ByP1 := MayBeOccupiedBy(ThisPlayer, TestX, TestY);
            ByP2 := MayBeOccupiedBy(ThisPlayer xor 1, TestX, TestY);
            if FieldPtr[TestX, TestY]^.PossibleOwners = OWNER_UNDECIDED
            then begin
              if ByP1 and (not ByP2) then inc(Result);
              if (not ByP1) and ByP2 then dec(Result);
            end;
          end;

    for i := 1 to LINES_EXTENDING do PND[i]^ := ND[i];
  end; {with}
end;

function TPlayer.DetermineMove: TMove;
begin
  case FAILevel of
    0: Result := DetermineRandomMove;
    1: Result := DetermineRandomBestDefensiveMove;
    2: Result := DetermineRandomBestOffensiveMove;
    3: Result := DetermineRandomBestAverageMove;
    4: Result := DetermineBestMiniMaxMove;
    5: Result := DetermineBestAlphaBetaMove;
  end;
end;

function TPlayer.DetermineRandomMove: TMove;
var Move: TMove;
    ResultArray: array of TMove;
    ResultArraySize: integer;
begin
  SetLength(ResultArray, 32);
  ResultArraySize := 0;
  with Move do begin
    X      := 0;       // da dieser Move nie mglich ist (sein sollte),
    Y      := 0;       // funktioniert es auch mit genau dieser Initialisierung
    Dir    := DIR_H;   // (siehe anschlieender Aufruf von NextMove)
    Player := ThisPlayer;
  end;
  while NextMove(Move) do begin
    inc(ResultArraySize);
    if (ResultArraySize > length(ResultArray)) then
      SetLength(ResultArray, ResultArraySize+32);
    ResultArray[ResultArraySize - 1] := Move;
  end;
  Result := ResultArray[random(ResultArraySize)];
  Finalize(ResultArray);
end;

function TPlayer.DetermineRandomBestDefensiveMove: TMove;
var SafedNodesNow,
    SafedNodesMax,
    ResultArraySize : integer;
    Move            : TMove;
    ResultArray     : array of TMove;
begin
  SetLength(ResultArray, 8);
  SafedNodesMax := -1;
  ResultArraySize := 0;
  with Move do begin
    X      := 0;       // da dieser Move nie mglich ist (sein sollte),
    Y      := 0;       // funktioniert es auch mit genau dieser Initialisierung
    Dir    := DIR_H;   // (siehe anschlieender Aufruf von NextMove)
    Player := ThisPlayer;
  end;
  while NextMove(Move) do begin
    SafedNodesNow := TheGame.Player[ThisPlayer].DetermineSafedNodes(Move);
    if SafedNodesNow = SafedNodesMax then begin
      inc(ResultArraySize);
      if (ResultArraySize > length(ResultArray)) then
        SetLength(ResultArray, ResultArraySize+8);
      ResultArray[ResultArraySize - 1] := Move;
    end else if SafedNodesNow > SafedNodesMax then begin
      SafedNodesMax := SafedNodesNow;
      ResultArraySize := 1;
      ResultArray[0] := Move;
    end;
  end;
  Result := ResultArray[random(ResultArraySize)];
  Finalize(ResultArray);
end;

function TPlayer.DetermineRandomBestOffensiveMove: TMove;
var BlockedMovesNow,
    BlockedMovesMax,
    ResultArraySize : integer;
    Move            : TMove;
    ResultArray     : array of TMove;
begin
  SetLength(ResultArray, 8);
  BlockedMovesMax := 0;
  ResultArraySize := 0;
  with Move do begin
    X      := 0;       // da dieser Move nie mglich ist (sein sollte),
    Y      := 0;       // funktioniert es auch mit genau dieser Initialisierung
    Dir    := DIR_H;   // (siehe anschlieender Aufruf von NextMove)
    Player := ThisPlayer;
  end;
  while NextMove(Move) do begin
    BlockedMovesNow := TheGame.Player[ThisPlayer xor 1].
                       DetermineBlockedMoves(Move, []);
    if BlockedMovesNow = BlockedMovesMax then begin
      inc(ResultArraySize);
      if (ResultArraySize > length(ResultArray)) then
        SetLength(ResultArray, ResultArraySize+8);
      ResultArray[ResultArraySize - 1] := Move;
    end else if BlockedMovesNow > BlockedMovesMax then begin
      BlockedMovesMax := BlockedMovesNow;
      ResultArraySize := 1;
      ResultArray[0] := Move;
    end;
  end;
  Result := ResultArray[random(ResultArraySize)];
  Finalize(ResultArray);
end;

function TPlayer.DetermineRandomBestAverageMove: TMove;
const factor = 4;
var BlockedMovesNow,
    SafedNodesNow,
    GlobMax,
    ResultArraySize : integer;
    Move            : TMove;
    ResultArray     : array of TMove;
begin
  SetLength(ResultArray, 8);
  GlobMax := 0;
  ResultArraySize := 0;
  with Move do begin
    X      := 0;       // da dieser Move nie mglich ist (sein sollte),
    Y      := 0;       // funktioniert es auch mit genau dieser Initialisierung
    Dir    := DIR_H;   // (siehe anschlieender Aufruf von NextMove)
    Player := ThisPlayer;
  end;
  while NextMove(Move) do begin
    SafedNodesNow := TheGame.Player[ThisPlayer].DetermineSafedNodes(Move);
    BlockedMovesNow := TheGame.Player[ThisPlayer xor 1].
                       DetermineBlockedMoves(Move, []);
    if BlockedMovesNow + factor*SafedNodesNow = GlobMax then begin
      inc(ResultArraySize);
      if (ResultArraySize > length(ResultArray)) then
        SetLength(ResultArray, ResultArraySize+8);
      ResultArray[ResultArraySize - 1] := Move;
    end else if BlockedMovesNow + factor*SafedNodesNow > GlobMax then begin
      GlobMax := BlockedMovesNow + factor*SafedNodesNow ;
      ResultArraySize := 1;
      ResultArray[0] := Move;
    end;
  end;
  Result := ResultArray[random(ResultArraySize)];
  Finalize(ResultArray);
end;

function TPlayer.DetermineBestMiniMaxMove: TMove;
var Count, Min, MinNow, i: integer;
    RatedMoves: TRatedMoves;
begin
  AbortMove := false;
  MainForm.Lock;
  Min := INFIMUM_POS;
  DetermineBestRatedMoves(Count, RatedMoves);
  ProgressForm.ProgressBar.Position := 0;
  ProgressForm.ProgressBar.Max      :=
               Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH));
  ProgressForm.Show;
  for i := 1 to Count do begin
    TheGame.Push;
    // wir sind Minimierer (wir minimieren die Anzahl noch mglicher Moves
    // fr den Gegner)
    MinNow := Minimize(ThisPlayer xor 1, RatedMoves[i].Move, 1);
    if MinNow < Min then begin
      Min    := MinNow;
      Result := RatedMoves[i].Move;
    end;
    TheGame.Pop;
    if AbortMove then break;
  end;
  ProgressForm.Hide;
  MainForm.Unlock;
end;

function TPlayer.DetermineBestAlphaBetaMove: TMove;
var Count, Alpha, Beta, BetaNow, i: integer;
    RatedMoves: TRatedMoves;
begin
  AbortMove := false;
  MainForm.Lock;
  Alpha := INFIMUM_NEG;
  Beta  := INFIMUM_POS;
  DetermineBestRatedMoves(Count, RatedMoves);
  ProgressForm.ProgressBar.Position := 0;
  ProgressForm.ProgressBar.Max      :=
               Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH));
  ProgressForm.Show;
  for i := 1 to Count do begin
    TheGame.Push;
    // wir sind Minimierer (wir minimieren die Anzahl noch mglicher Moves
    // fr den Gegner)
    BetaNow := Minimize(ThisPlayer xor 1, RatedMoves[i].Move, 1,
                        Alpha, Beta);
    if BetaNow < Beta then begin
      Beta   := BetaNow;
      Result := RatedMoves[i].Move;
    end;
    TheGame.Pop;
    if AbortMove then break;
  end;
  ProgressForm.Hide;
  MainForm.Unlock;
end;

function TPlayer.CanMove: Boolean;
begin
  Result := FRemainingMoves > 0;
end;

procedure TPlayer.DetermineBestRatedMoves(var Count: integer;
                                          var RatedMoves: TRatedMoves);
var Move : TMove;
    RatedMove: TRatedMove;
    i: integer;
begin
  Count := 0;
  with Move do begin
    X      := 0;       // da dieser Move nie mglich ist (sein sollte),
    Y      := 0;       // funktioniert es auch mit genau dieser Initialisierung
    Dir    := DIR_H;   // (siehe anschlieender Aufruf von NextMove)
    Player := ThisPlayer;
  end;
  while NextMove(Move) do begin
    RatedMove.Rating := TheGame.Player[ThisPlayer xor 1]
                               .DetermineBlockedMoves(Move, []);
    RatedMove.Move   := Move;
    if Count < GAMETREE_BRANCHFACTOR then begin
      inc(Count);
      RatedMoves[Count] := RatedMove;
    end;
    for i := (Count - 1) downto 1 do
      if RatedMove.Rating > RatedMoves[i].Rating then begin
        RatedMoves[i+1] := RatedMoves[i];
        RatedMoves[i]   := RatedMove;
      end else break;
  end;
end;

end.